import ctypes
import unittest
from collections import namedtuple
from multiprocessing import Process, Queue
from queue import Empty
from threading import Thread

import pdb

SerializableMutationTestResult = namedtuple(
	'SerializableMutationTestResult', [
		'is_incompetent',
		'is_survived',
		'killer',
		'exception_traceback',
		'exception',
		'tests_run',
		'analysis_data'
	]
)


class MutationTestResult(unittest.TestResult):
	def __init__(self, *args, coverage_injector=None, **kwargs):
		super(MutationTestResult, self).__init__(*args, **kwargs)
		self.type_error = None
		self.failfast = True
		self.coverage_injector = coverage_injector

	def addError(self, test, err):
		if err[0] == TypeError:
			self.type_error = err
		else:
			super(MutationTestResult, self).addError(test, err)

	def is_incompetent(self):
		return bool(self.type_error)

	def is_survived(self):
		return self.wasSuccessful()

	def get_killer(self):
		if self.failures:
			return self.failures[0][0]
		elif self.errors:
			return self.errors[0][0]

	def get_exception_traceback(self):
		if self.failures:
			return self.failures[0][1]
		elif self.errors:
			return self.errors[0][1]

	def get_exception(self):
		if self.type_error:
			return self.type_error[1]

	def get_analysis_data(self):
		if hasattr(self,"analysis_data"):
			return self.analysis_data
		return None
	def set_analysis_data(self,analysis_data):
		self.analysis_data = analysis_data
		
	def serialize(self):
		return SerializableMutationTestResult(
			self.is_incompetent(),
			self.is_survived(),
			str(self.get_killer()),
			str(self.get_exception_traceback()),
			self.get_exception(),
			self.testsRun - len(self.skipped),
			self.get_analysis_data()
		)
		
class MutationTestRunner:
	def __init__(self, suite):
		super().__init__()
		self.suite = suite

	def run(self):
		try:
			separate_test_suites = {x._testMethodName:unittest.suite.TestSuite([x]) for x in self.suite._tests[0]._tests}
		except IndexError:
			analysis_data = MutantMatrixGenerator.empty
			with open('x.txt','w') as f:
				f.write(str(self.suite._tests))
		else:	
			analysis_data = {x._testMethodName:None for x in self.suite._tests[0]._tests}
			for testMethodName in analysis_data.keys():
				result = MutationTestResult()
				separate_test_suites[testMethodName].run(result)
				analysis_data[testMethodName] = result.serialize()
			
		result = MutationTestResult()
		self.suite.run(result)
		result.analysis_data = analysis_data
		self.set_result(result)
		
class MutationTestRunnerProcess(MutationTestRunner, Process):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.queue = Queue()

	def get_result(self, live_time):
		try:
			return self.queue.get(timeout=live_time)
		except Empty:
			return None

	def set_result(self, result):
		self.queue.put_nowait(result.serialize())


class MutationTestRunnerThread(MutationTestRunner, Thread):
	daemon = True

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.result = None

	def terminate(self):
		if self.isAlive():
			res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self.ident), ctypes.py_object(SystemExit))
			if res == 0:
				raise ValueError('Invalid thread id.')
			elif res != 1:
				raise SystemError('Thread killing failed.')

	def set_result(self, result):
		self.result = result

	def get_result(self, live_time):
		self.join(live_time)
		if self.is_alive():
			return None
		return self.result.serialize()
		
class MutantMatrixGenerator:
	incompetence = 'incompetence'
	empty = 'empty'
	def __init__(self):
		self.data_list = {}
	def add_mutant(self,number,data):
		self.data_list[number] = data
	def generate_raport(self,fileName='raport.txt'):
		with open(fileName,'w') as f:
			f.write("mutant no || killed || survived || incompetent || timeout\n")
			f.write("----------++--------++----------++-------------++--------\n")
			digits = len(str(len(self.data_list)))
			format_string = (" "*(9-digits))+"{0:0"+str(digits)+"d} ||    {1}   ||     {2}    ||      {3}      ||    {4}\n"
			for key,value in self.data_list.items():
				incompetence = False
				survived = False
				killed = False
				timeout = False
				if value == MutantMatrixGenerator.empty:
					pass
				elif value is None:
					timeout = True
				elif value == MutantMatrixGenerator.incompetence or value.is_incompetent:
					incompetence = True
				elif value.is_survived:
					survived = True
				else:
					killed = True
				f.write(format_string.format(key,"*" if killed else " ","*" if survived else " ","*" if incompetence else " ","*" if timeout else " "))
	def generate_matrix(self,fileName = 'matrix.txt'):
		all_tests = {}
		for entry in self.data_list.values():
			if not isinstance(entry,str):
				all_tests|=entry.analysis_data.keys()
		test_names = list(all_tests)
		
		digits = len(str(len(self.data_list)))
		format_string = (" "*(9-digits))+"{0:0"+str(digits)+"d} "
		i=1
		title_string = "mutant no "
		
		for test_name in test_names:
			title_string += "|| "+test_name+" "
			l = len(test_name)
			#format_string += "|| "+(" "*(l//2+l%2))+"{"+str(i)+"}" + (" "*(l//2-1))+" "
			format_string +="|| {"+str(i)+"} "
			i+=1
		else:
			title_string+="\n"
			format_string+="\n"
			
		legend = "\nx - no test\ne - empty result\ni - incompetent\nk - killed\ns - survived\nt - timeout\n"
		with open(fileName,'w') as f:
			f.write(title_string)
			for key,value in self.data_list.items():
			
				test_results = []
				for test_name in test_names:
					if not test_name in value.analysis_data:
						test_results.append('x')
					elif value.analysis_data[test_name] is None:
						test_results.append('t')
					elif value.analysis_data[test_name] == MutantMatrixGenerator.empty:
						test_results.append('e')
					elif value == MutantMatrixGenerator.incompetence or value.analysis_data[test_name].is_incompetent:
						test_results.append('i')
					elif value.analysis_data[test_name].is_survived:
						test_results.append('s')
					else:
						test_results.append('k')
				f.write(format_string.format(key,*test_results))
			f.write(legend)